@@ -1,4 +1,4 @@
-# $Id: Canonical.pm 43961 2015-10-05 11:54:50Z wsl $
+# $Id: Canonical.pm 44525 2016-01-13 12:48:24Z wsl $
 # $URL: https://svn.uvt.nl/its-id/trunk/sources/libnet-ldap-schema-canonical-perl/lib/Net/LDAP/Schema/Canonical.pm $
 
 use strict;
@@ -8,8 +8,10 @@ package Net::LDAP::Schema;
 
 use Unicode::Normalize;
 use Encode;
+use POSIX qw(mktime floor);
 
-my %rfc4519_namemap = (
+# RFC 4519
+my %rfc_namemap = (
 	'2.5.4.15' => 'businessCategory',
 	'2.5.4.6' => 'c',
 	'2.5.4.3' => 'cn',
@@ -107,71 +109,193 @@ my %rfc4519_namemap = (
 	x500uniqueidentifier => 'x500UniqueIdentifier',
 );
 
-my %rfc4519_typemap = (
-	businessCategory => 'caseIgnoreMatch',
-	c => 'caseIgnoreMatch',
-	cn => 'caseIgnoreMatch',
-	dc => 'caseIgnoreIA5Match',
-	description => 'caseIgnoreMatch',
-	destinationIndicator => 'caseIgnoreMatch',
-	distinguishedName => 'distinguishedNameMatch',
-	dnQualifier => 'caseIgnoreMatch',
-	enhancedSearchGuide => '',
-	facsimileTelephoneNumber => '',
-	generationQualifier => 'caseIgnoreMatch',
-	givenName => 'caseIgnoreMatch',
-	houseIdentifier => 'caseIgnoreMatch',
-	initials => 'caseIgnoreMatch',
-	internationalISDNNumber => 'numericStringMatch',
-	l => 'caseIgnoreMatch',
-	member => 'distinguishedNameMatch',
-	name => 'caseIgnoreMatch',
-	o => 'caseIgnoreMatch',
-	ou => 'caseIgnoreMatch',
-	owner => 'distinguishedNameMatch',
-	physicalDeliveryOfficeName => 'caseIgnoreMatch',
-	postalAddress => 'caseIgnoreListMatch',
-	postalCode => 'caseIgnoreMatch',
-	postOfficeBox => 'caseIgnoreMatch',
-	preferredDeliveryMethod => '',
-	registeredAddress => 'caseIgnoreListMatch',
-	roleOccupant => 'distinguishedNameMatch',
-	searchGuide => '',
-	seeAlso => 'distinguishedNameMatch',
-	serialNumber => 'caseIgnoreMatch',
-	sn => 'caseIgnoreMatch',
-	st => 'caseIgnoreMatch',
-	street => 'caseIgnoreMatch',
-	telephoneNumber => 'telephoneNumberMatch',
-	teletexTerminalIdentifier => '',
-	telexNumber => '',
-	title => 'caseIgnoreMatch',
-	uid => 'caseIgnoreMatch',
-	uniqueMember => 'uniqueMemberMatch',
-	userPassword => 'octetStringMatch',
-	x121Address => 'numericStringMatch',
-	x500UniqueIdentifier => 'bitStringMatch',
+my %rfc_typemap = (
+	# attribute identifiers from RFC 4519
+	businesscategory => 'canonicalize_caseIgnore',
+	c => 'canonicalize_caseIgnore',
+	cn => 'canonicalize_caseIgnore',
+	dc => 'canonicalize_caseIgnore',
+	description => 'canonicalize_caseIgnore',
+	destinationindicator => 'canonicalize_caseIgnore',
+	distinguishedname => 'canonicalize_dn',
+	dnqualifier => 'canonicalize_caseIgnore',
+	enhancedsearchguide => 'canonicalize_enhancedGuide',
+	facsimiletelephonenumber => 'canonicalize_facsimileTelephoneNumber',
+	generationqualifier => 'canonicalize_caseIgnore',
+	givenname => 'canonicalize_caseIgnore',
+	houseidentifier => 'canonicalize_caseIgnore',
+	initials => 'canonicalize_caseIgnore',
+	internationalisdnnumber => 'canonicalize_numericString',
+	l => 'canonicalize_caseIgnore',
+	member => 'canonicalize_dn',
+	name => 'canonicalize_caseIgnore',
+	o => 'canonicalize_caseIgnore',
+	ou => 'canonicalize_caseIgnore',
+	owner => 'canonicalize_dn',
+	physicaldeliveryofficename => 'canonicalize_caseIgnore',
+	postaladdress => 'canonicalize_caseIgnoreList',
+	postalcode => 'canonicalize_caseIgnore',
+	postofficebox => 'canonicalize_caseIgnore',
+	preferreddeliverymethod => 'canonicalize_deliveryMethod',
+	registeredaddress => 'canonicalize_caseIgnoreList',
+	roleoccupant => 'canonicalize_dn',
+	searchguide => 'canonicalize_searchGuide',
+	seealso => 'canonicalize_dn',
+	serialnumber => 'canonicalize_caseIgnore',
+	sn => 'canonicalize_caseIgnore',
+	st => 'canonicalize_caseIgnore',
+	street => 'canonicalize_caseIgnore',
+	telephonenumber => 'canonicalize_telephoneNumber',
+	teletexterminalidentifier => 'canonicalize_teletexTerminalIdentifier',
+	telexnumber => 'canonicalize_telexNumber',
+	title => 'canonicalize_caseIgnore',
+	uid => 'canonicalize_caseIgnore',
+	uniquemember => 'canonicalize_nameAndOptionalUID',
+	userpassword => 'canonicalize_octetString',
+	x121address => 'canonicalize_numericString',
+	x500uniqueidentifier => 'canonicalize_bitString',
 
-	# FIXME: add all syntax identifiers from section 3.3
-	# FIXME: add all *Match identifiers from section 4.2
+	# syntax identifiers from RFC 4517 section 3.3
+	bitstring => 'canonicalize_bitString',
+	'1.3.6.1.4.1.1466.115.121.1.6' => 'canonicalize_bitString',
+	boolean => 'canonicalize_boolean',
+	'1.3.6.1.4.1.1466.115.121.1.7' => 'canonicalize_boolean',
+	countrystring => 'canonicalize_caseIgnore',
+	'1.3.6.1.4.1.1466.115.121.1.11' => 'canonicalize_caseIgnore',
+	deliverymethod => 'canonicalize_deliveryMethod',
+	'1.3.6.1.4.1.1466.115.121.1.14' => 'canonicalize_deliveryMethod',
+	directorystring => 'canonicalize_caseIgnore',
+	'1.3.6.1.4.1.1466.115.121.1.15' => 'canonicalize_caseIgnore',
+	ditcontentruledescription => 'canonicalize_ditContentRuleDescription',
+	'1.3.6.1.4.1.1466.115.121.1.16' => 'canonicalize_DITContentRuleDescription',
+	ditstructureruledescription => 'canonicalize_ditStructureRuleDescription',
+	'1.3.6.1.4.1.1466.115.121.1.17' => 'canonicalize_ditStructureRuleDescription',
+	dn => 'canonicalize_dn',
+	'1.3.6.1.4.1.1466.115.121.1.12' => 'canonicalize_dn',
+	enhancedguide => 'canonicalize_enhancedGuide',
+	'1.3.6.1.4.1.1466.115.121.1.21' => 'canonicalize_enhancedGuide',
+	facsimiletelephonenumber => 'canonicalize_facsimileTelephoneNumber',
+	'1.3.6.1.4.1.1466.115.121.1.22' => 'canonicalize_facsimileTelephoneNumber',
+	fax => 'canonicalize_fax',
+	'1.3.6.1.4.1.1466.115.121.1.23' => 'canonicalize_fax',
+	generalizedtime => 'canonicalize_generalizedTime',
+	'1.3.6.1.4.1.1466.115.121.1.24' => 'canonicalize_generalizedTime',,
+	guide => 'canonicalize_guide',
+	'1.3.6.1.4.1.1466.115.121.1.25' => 'canonicalize_guide',
+	ia5string => 'canonicalize_caseIgnore',
+	'1.3.6.1.4.1.1466.115.121.1.26' => 'canonicalize_caseIgnore',
+	integer => 'canonicalize_integer',
+	'1.3.6.1.4.1.1466.115.121.1.27' => 'canonicalize_integer',
+	jpeg => 'canonicalize_jpeg',
+	'1.3.6.1.4.1.1466.115.121.1.28' => 'canonicalize_jpeg',
+	ldapsyntaxdescription => 'canonicalize_ldapSyntaxDescription',
+	'1.3.6.1.4.1.1466.115.121.1.54' => 'canonicalize_ldapSyntaxDescription',
+	matchingruledescription => 'canonicalize_matchingRuleDescription',
+	'1.3.6.1.4.1.1466.115.121.1.30' => 'canonicalize_matchingRuleDescription',
+	matchingruleusedescription => 'canonicalize_matchingRuleUseDescription',
+	'1.3.6.1.4.1.1466.115.121.1.31' => 'canonicalize_matchingRuleUseDescription',
+	nameandoptionaluid => 'canonicalize_nameAndOptionalUID',
+	'1.3.6.1.4.1.1466.115.121.1.34' => 'canonicalize_nameAndOptionalUID',
+	nameformdescription => 'canonicalize_nameFormDescription',
+	'1.3.6.1.4.1.1466.115.121.1.35' => 'canonicalize_nameFormDescription',
+	numericstring => 'canonicalize_numericString',
+	'1.3.6.1.4.1.1466.115.121.1.36' => 'canonicalize_numericString',
+	objectclassdescription => 'canonicalize_objectClassDescription',
+	'1.3.6.1.4.1.1466.115.121.1.3' => 'canonicalize_objectClassDescription',
+	octetstring => 'canonicalize_octetString',
+	'1.3.6.1.4.1.1466.115.121.1.40' => 'canonicalize_octetString',
+	oid => 'canonicalize_oid',
+	'1.3.6.1.4.1.1466.115.121.1.38' => 'canonicalize_oid',
+	othermailbox => 'canonicalize_otherMailbox',
+	'1.3.6.1.4.1.1466.115.121.1.39' => 'canonicalize_otherMailbox',
+	postaladdress => 'canonicalize_PostalAddress',
+	'1.3.6.1.4.1.1466.115.121.1.41' => 'canonicalize_PostalAddress',
+	printablestring => 'canonicalize_caseExact',
+	'1.3.6.1.4.1.1466.115.121.1.44' => 'canonicalize_caseExact',
+#	"The Substring Assertion syntax is used only as the syntax of
+#	assertion values in the extensible match.  It is not used as an
+#	attribute syntax, or in the SubstringFilter [RFC4511]."
+#	substringassertion => '',
+#	'1.3.6.1.4.1.1466.115.121.1.58' => '',
+	telephonenumber => 'canonicalize_telephoneNumber',
+	'1.3.6.1.4.1.1466.115.121.1.50' => 'canonicalize_telephoneNumber',
+	teletexterminalidentifier => 'canonicalize_teletexTerminalIdentifier',
+	'1.3.6.1.4.1.1466.115.121.1.51' => 'canonicalize_teletexTerminalIdentifier',
+	telexnumber => 'canonicalize_telexNumber',
+	'1.3.6.1.4.1.1466.115.121.1.52' => 'canonicalize_telexNumber',
+	utctime => 'canonicalize_utcTime',
+	'1.3.6.1.4.1.1466.115.121.1.53' => 'canonicalize_utcTime',
+
+	# *Match identifiers from RFC 4517 section 4.2
+	bitstringmatch => 'canonicalize_bitString',
+	'2.5.13.16' => 'canonicalize_bitString',
+	booleanmatch => 'canonicalize_boolean',
+	'2.5.13.13' => 'canonicalize_boolean',
+	caseexactia5match => 'canonicalize_caseExact',
+	'1.3.6.1.4.1.1466.109.114.1' => 'canonicalize_caseExact',
+	caseexactmatch => 'canonicalize_caseExact',
+	'2.5.13.5' => 'canonicalize_caseExact',
+	caseexactorderingmatch => 'canonicalize_caseExact',
+	'2.5.13.6' => 'canonicalize_caseExact',
+	caseexactsubstringsmatch => 'canonicalize_caseExact',
+	'2.5.13.7' => 'canonicalize_caseExact',
+	caseignoreia5match => 'canonicalize_caseIgnore',
+	'1.3.6.1.4.1.1466.109.114.2' => 'canonicalize_caseIgnore',
+	caseignoreia5substringsmatch => 'canonicalize_caseIgnore',
+	'1.3.6.1.4.1.1466.109.114.3' => 'canonicalize_caseIgnore',
+	caseignorelistmatch => 'canonicalize_caseIgnoreList',
+	'2.5.13.11' => 'canonicalize_caseIgnoreList',
+	caseignorelistsubstringsmatch => 'canonicalize_caseIgnoreList',
+	'2.5.13.12' => 'canonicalize_caseIgnoreList',
+	caseignorematch => 'canonicalize_caseIgnore',
+	'2.5.13.2' => 'canonicalize_caseIgnore',
+	caseignoreorderingmatch => 'canonicalize_caseIgnore',
+	'2.5.13.3' => 'canonicalize_caseIgnore',
+	caseignoresubstringsmatch => 'canonicalize_caseIgnore',
+	'2.5.13.4' => 'canonicalize_caseIgnore',
+	directorystringfirstcomponentmatch => 'canonicalize_caseIgnore',
+	'2.5.13.31' => 'canonicalize_caseIgnore',
+	distinguishednamematch => 'canonicalize_dn',
+	'2.5.13.1' => 'canonicalize_dn',
+	generalizedtimematch => 'canonicalize_generalizedTime',
+	'2.5.13.27' => 'canonicalize_generalizedTime',
+	generalizedtimeorderingmatch => 'canonicalize_generalizedTime',
+	'2.5.13.28' => 'canonicalize_generalizedTime',
+	integerfirstcomponentmatch => 'canonicalize_integer',
+	'2.5.13.29' => 'canonicalize_integer',
+	integermatch => 'canonicalize_integer',
+	'2.5.13.14' => 'canonicalize_integer',
+	integerorderingmatch => 'canonicalize_integer',
+	'2.5.13.15' => 'canonicalize_integer',
+	keywordmatch => 'canonicalize_caseIgnore',
+	'2.5.13.33' => 'canonicalize_caseIgnore',
+	numericstringmatch => 'canonicalize_numericString',
+	'2.5.13.8' => 'canonicalize_numericString',
+	numericstringorderingmatch => 'canonicalize_numericString',
+	'2.5.13.9' => 'canonicalize_numericString',
+	numericstringsubstringsmatch => 'canonicalize_numericString',
+	'2.5.13.10' => 'canonicalize_numericString',
+	objectidentifierfirstcomponentmatch => 'canonicalize_objectIdentifier',
+	'2.5.13.30' => 'canonicalize_objectIdentifier',
+	objectidentifiermatch => 'canonicalize_objectIdentifier',
+	'2.5.13.0' => 'canonicalize_objectIdentifier',
+	octetstringmatch => 'canonicalize_octetString',
+	'2.5.13.17' => 'canonicalize_octetString',
+	octetstringorderingmatch => 'canonicalize_octetString',
+	'2.5.13.18' => 'canonicalize_octetString',
+	telephonenumbermatch => 'canonicalize_telephoneNumber',
+	'2.5.13.20' => 'canonicalize_telephoneNumber',
+	telephonenumbersubstringsmatch => 'canonicalize_telephoneNumber',
+	'2.5.13.21' => 'canonicalize_telephoneNumber',
+	uniquemembermatch => 'canonicalize_nameAndOptionalUID',
+	'2.5.13.23' => 'canonicalize_nameAndOptionalUID',
+	wordmatch => 'canonicalize_caseIgnore',
+	'2.5.13.32' => 'canonicalize_caseIgnore',
 );
 
-sub canonicalize_name {
-	my ($self, $name) = @_;
-
-	$name = lc($name);
-	$name =~ s/(;.*)//;
-	my $options = $1 // '';
-
-	return "$rfc4519_namemap{$name}$options"
-		if exists $rfc4519_namemap{$name};
-
-	if(my $c = ref $self && $self->attribute($name)) {
-		return $c->{name}.$options;
-	} else {
-		return $name.$options;
-	}
-}
+#while(my ($key, $value) = each %rfc_namemap) {
+#	$rfc_typemap{$key} //= $rfc_typemap{lc($value)};
+#}
 
 # RFC 3454 appendix B.2
 my %rfc3454_casemap = (
@@ -1549,30 +1673,19 @@ my %rfc3454_casemap = (
 );
 my ($rfc3454_casemap) = map { qr{[$_]} } join('', keys %rfc3454_casemap);
 
-sub canonicalize_value {
-	my ($self, $name, $value) = @_;
-	my $type = $rfc4519_typemap{$name}
-		|| (ref $self && $self->matchingrule_for_attribute($name, 'equality'))
-		|| $name;
+# See RFC 4518 section 2 for a description of the steps below
 
-    return $self->canonicalize_dn($value) 
-        if $type eq 'distinguishedNameMatch';
+# 2.1.  Transcode
+sub canonicalize_stringprep_transcode {
+	my ($self, $value) = @_;
+	utf8::decode($value) or die "invalid UTF-8 sequence in '$value'\n"
+		unless utf8::is_utf8($value);
+	return $value;
+}
 
-	return $value
-		unless $type =~ /^(?:case(?:Exact|Ignore)|numericStringMatch\z|integerMatch\z|telephoneNumberMatch\z)/a;
-
-	# TODO:
-	# bit string        https://tools.ietf.org/html/rfc4517#section-3.3.2
-	# boolean           https://tools.ietf.org/html/rfc4517#section-3.3.3
-	# fax numbers       https://tools.ietf.org/html/rfc4517#section-3.3.11
-	# generalized time  https://tools.ietf.org/html/rfc4517#section-3.3.13
-
-	# See RFC 4518 section 2 for a description of the steps below
-
-	# 2.1.  Transcode
-	utf8::decode($value) or die "invalid UTF-8 sequence in '$value'\n";
-
-	# 2.2.  Map
+# 2.2.  Map
+sub canonicalize_stringprep_map {
+	my ($self, $value) = @_;
 	# map to nothing
 	$value =~ s/[
 		\x{00AD}
@@ -1618,46 +1731,423 @@ sub canonicalize_value {
 		\x{3000}
 	]+/\x{0020}/agx;
 
-	$value =~ s{($rfc3454_casemap)}{$rfc3454_casemap{$1}}ego
-		if $type =~ /^caseIgnore/ || $type eq 'numericStringMatch';
+	return $value;
+}
 
-	# 2.3.  Normalize
+sub canonicalize_stringprep_casemap {
+	my ($self, $value) = @_;
+	$value =~ s{($rfc3454_casemap)}{$rfc3454_casemap{$1}}ego;
+	return $value;
+}
+
+# 2.3.  Normalize
+sub canonicalize_stringprep_normalize {
+	my ($self, $value) = @_;
 	$value = NFKC($value);
+	return $value;
+}
 
-	# 2.4.  Prohibit
-	# FIXME
+# 2.4.  Prohibit
+sub canonicalize_stringprep_prohibit {
+	my ($self, $value) = @_;
+	# (no-op, this implementation is not intended to be validating)
+	return $value;
+}
 
-	# 2.5.  Check bidi
+# 2.5.  Check bidi
+sub canonicalize_stringprep_bidi {
+	my ($self, $value) = @_;
 	# (no-op)
+	return $value;
+}
 
-	# 2.6.  Insignificant Character Handling
-	if($type =~ /^case(?:Exact|Ignore)/) {
-		$value =~ s/^ (?!\p{M})//g;
-		$value =~ s/ (?!\p{M})\z//g;
-#		$value = " $value ";
-	} elsif($type eq 'numericStringMatch') {
-		$value =~ s/\x{0020}(?!\p{M})//g;
-	} elsif($type eq 'integerMatch') {
-		$value =~ s/\x{0020}+//g;
-		$value =~ s/^(-?)0*([1-9][0-9]*|0)\z/$1$2/a
-			or die "invalid integer value '$value'\n";
-	} elsif($type eq 'telephoneNumberMatch') {
-		$value =~ s/[
-			\x{0020}
-			\x{002D}
-			\x{058A}
-			\x{2010}
-			\x{2011}
-			\x{2212}
-			\x{FE63}
-			\x{FF0D}
-		](?!\p{M})//agx;
-	}
-
+sub canonicalize_caseIgnore {
+	my ($self, $value) = @_;
+	$value = $self->canonicalize_stringprep_transcode($value);
+	$value = $self->canonicalize_stringprep_map($value);
+	$value = $self->canonicalize_stringprep_casemap($value);
+	$value = $self->canonicalize_stringprep_normalize($value);
+#	$value = $self->canonicalize_stringprep_prohibit($value);
+#	$value = $self->canonicalize_stringprep_bidi($value);
+	$value =~ s/^ (?!\p{M})//g;
+	$value =~ s/ (?!\p{M})\z//g;
 	utf8::encode($value);
 	return $value;
 }
 
+sub canonicalize_caseExact {
+	my ($self, $value) = @_;
+	$value = $self->canonicalize_stringprep_transcode($value);
+	$value = $self->canonicalize_stringprep_map($value);
+	$value = $self->canonicalize_stringprep_normalize($value);
+#	$value = $self->canonicalize_stringprep_prohibit($value);
+#	$value = $self->canonicalize_stringprep_bidi($value);
+	$value =~ s/^ (?!\p{M})//g;
+	$value =~ s/ (?!\p{M})\z//g;
+	utf8::encode($value);
+	return $value;
+}
+
+sub canonicalize_numericString {
+	my ($self, $value) = @_;
+	$value = $self->canonicalize_stringprep_transcode($value);
+	$value = $self->canonicalize_stringprep_map($value);
+	$value = $self->canonicalize_stringprep_normalize($value);
+#	$value = $self->canonicalize_stringprep_prohibit($value);
+#	$value = $self->canonicalize_stringprep_bidi($value);
+	$value =~ s/\x{0020}(?!\p{M})//g;
+	utf8::encode($value);
+	return $value;
+}
+
+sub canonicalize_integer {
+	my ($self, $value) = @_;
+	$value = $self->canonicalize_stringprep_transcode($value);
+	$value = $self->canonicalize_stringprep_map($value);
+	$value = $self->canonicalize_stringprep_normalize($value);
+#	$value = $self->canonicalize_stringprep_prohibit($value);
+#	$value = $self->canonicalize_stringprep_bidi($value);
+	$value =~ s/\x{0020}+//g;
+	$value =~ s/^(-?)0*([1-9][0-9]*|0)\z/$1$2/a
+		or die "invalid integer value '$value'\n";
+	utf8::encode($value);
+	return $value;
+}
+
+sub canonicalize_telephoneNumber {
+	my ($self, $value) = @_;
+	$value = $self->canonicalize_stringprep_transcode($value);
+	$value = $self->canonicalize_stringprep_map($value);
+	$value = $self->canonicalize_stringprep_normalize($value);
+#	$value = $self->canonicalize_stringprep_prohibit($value);
+#	$value = $self->canonicalize_stringprep_bidi($value);
+	$value =~ s/[
+		\x{0020}
+		\x{002D}
+		\x{058A}
+		\x{2010}
+		\x{2011}
+		\x{2212}
+		\x{FE63}
+		\x{FF0D}
+	](?!\p{M})//agx;
+	utf8::encode($value);
+	return $value;
+}
+
+sub canonicalize_caseIgnoreList {
+	my ($self, $value) = @_;
+	my @els = split(/((?:[^\\\$]|\\[\\\$])+)/a, $value, -1);
+	die "invalid list syntax in '$value'\n" if @els < 2;
+	die "invalid list syntax in '$value'\n" if shift @els ne '';
+	die "invalid list syntax in '$value'\n" if pop @els ne '';
+	foreach(@els) {
+		next if $_ eq '$';
+		s/\\([\\\$])/$1/g;
+		$_ = $self->canonicalize_caseIgnore($_);
+		s/([\\\$])/\\$1/g;
+	}
+	$value = join('', @els);
+	utf8::encode($value);
+	return $value;
+}
+
+sub canonicalize_bitString {
+	my ($self, $value) = @_;
+	# https://tools.ietf.org/html/rfc4517#section-3.3.2
+	warn "canonicalize_bitString is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_boolean {
+	my ($self, $value) = @_;
+	# https://tools.ietf.org/html/rfc4517#section-3.3.3
+	$value = $self->canonicalize_caseExact($value);
+	$value = uc($1);
+	return $value;
+}
+
+sub canonicalize_facsimileTelephoneNumber {
+	my ($self, $value) = @_;
+	# https://tools.ietf.org/html/rfc4517#section-3.3.11
+	warn "canonicalize_facsimileTelephoneNumber is not implemented yet";
+	return $value;
+}
+
+# The slowest NIH bigint multiplication function in the world.
+#sub slow_stringmul {
+#	use integer;
+#	my ($str, $mul) = @_;
+#	my @str = reverse map { ord($_) - 48 } unpack('(a)*', $str);
+#	my $rem = 0;
+#	foreach(@str) {
+#		$rem += $_ * $mul;
+#		$_ = $rem % 10;
+#		$rem /= 10;
+#	}
+#	while($rem) {
+#		push @str, $rem % 10;
+#		$rem /= 10;
+#	}
+#	return join('', map { chr($_ + 48) } reverse @str);
+#}
+
+sub stringmul {
+	my ($str, $mul) = @_;
+	my $len = length($str);
+	my $res;
+	if($len > 9) {
+		unless(Math::BigInt->can("new")) {
+			local $@;
+			eval 'use Math::BigInt; 1'
+				or die $@;
+		}
+		my $big = new Math::BigInt($str || 0);
+		$big->bmul($mul);
+		$res = $big->bstr || '';
+	} else {
+		$res = $str * $mul || '';
+	}
+	# correct the padding
+	return '0'x($len - length($res)) . $res;
+}
+
+sub canonicalize_generalizedTime {
+	my ($self, $value) = @_;
+
+	# https://tools.ietf.org/html/rfc4517#section-3.3.13
+
+	my ($year, $month, $day, $hour, $minute, $second, $frac, $zonesign, $zonehour, $zonemin) = $value =~ /^
+		(\d{4}) (\d{2}) (\d{2})
+		(\d{2}) (?: (\d{2}) (\d{2})? )?
+		(?: [,.] (\d+) )?
+		(?: Z | ([+-]) (\d{2}) (\d{2})? )
+	\z/xa or die "unable to parse generalizedTime string '$value'\n";
+
+	my $leapsecond = ($second // 0) == 60;
+	$second = 59 if $leapsecond;
+
+	my $time = do {
+		local $ENV{TZ} = 'GMT';
+		mktime($second // 0, $minute // 0, $hour, $day, $month - 1, $year - 1900)
+	};
+
+	if($frac) {
+		$frac =~ s/0+\z//;
+		unless(defined $second) {
+			my $fraclen = length($frac);
+			my $fracsec = stringmul($frac, defined $minute ? 60 : 3600);
+			$frac = substr($fracsec, -$fraclen, $fraclen, '');
+			$frac =~ s/0+\z//;
+			$time += $fracsec if $fracsec;
+		}
+	}
+
+	if($zonesign) {
+		$zonemin //= 0;
+		$time -= int("$zonesign$zonehour") * 3600 + int("$zonesign$zonemin") * 60;
+	}
+
+	($second, $minute, $hour, $day, $month, $year) = gmtime($time);
+
+	$second = 60 if $leapsecond;
+
+	$value = sprintf('%04d%02d%02d%02d%02d%02d',
+		$year + 1900, $month + 1, $day,
+		$hour, $minute, $second);
+
+	$value .= ".$frac" if $frac;
+	$value .= 'Z';
+
+	return $value;
+}
+
+sub canonicalize_deliveryMethod {
+	my ($self, $value) = @_;
+	warn "canonicalize_deliveryMethod is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_ditContentRuleDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_ditContentRuleDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_DITContentRuleDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_DITContentRuleDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_ditStructureRuleDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_ditStructureRuleDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_enhancedGuide {
+	my ($self, $value) = @_;
+	warn "canonicalize_enhancedGuide is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_fax {
+	my ($self, $value) = @_;
+	warn "canonicalize_fax is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_guide {
+	my ($self, $value) = @_;
+	warn "canonicalize_guide is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_jpeg {
+	my ($self, $value) = @_;
+	warn "canonicalize_jpeg is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_ldapSyntaxDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_ldapSyntaxDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_matchingRuleDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_matchingRuleDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_matchingRuleUseDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_matchingRuleUseDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_nameAndOptionalUID {
+	my ($self, $value) = @_;
+	my $uid = $value =~ s/(\#.*)// ? $1 : '';
+	$value = $self->canonicalize_dn($value);
+	$value .= $uid;
+	return $value;
+}
+
+sub canonicalize_nameFormDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_nameFormDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_objectClassDescription {
+	my ($self, $value) = @_;
+	warn "canonicalize_objectClassDescription is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_objectIdentifier {
+	my ($self, $value) = @_;
+	warn "canonicalize_objectIdentifier is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_octetString {
+	my ($self, $value) = @_;
+	return $value;
+}
+
+sub canonicalize_oid {
+	my ($self, $value) = @_;
+	warn "canonicalize_oid is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_otherMailbox {
+	my ($self, $value) = @_;
+	warn "canonicalize_otherMailbox is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_PostalAddress {
+	my ($self, $value) = @_;
+	warn "canonicalize_PostalAddress is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_searchGuide {
+	my ($self, $value) = @_;
+	warn "canonicalize_searchGuide is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_teletexTerminalIdentifier {
+	my ($self, $value) = @_;
+	warn "canonicalize_teletexTerminalIdentifier is not implemented yet";
+	return $value;
+}
+
+sub canonicalize_name {
+	my ($self, $name) = @_;
+
+	$name = lc($name);
+	my $options = $name =~ s/(;.*)// ? $1 : '';
+
+	return $rfc_namemap{$name}.$options
+		if exists $rfc_namemap{$name};
+
+	if(my $c = ref $self && $self->attribute($name)) {
+		return $c->{name}.$options;
+	} else {
+		return $name.$options;
+	}
+}
+
+# Try various gambits to get a relevant canonicalization strategy identifier
+sub canonicalize_type {
+	my ($self, $name) = @_;
+
+	$name =~ s/;.*//;
+	$name = lc($self->canonicalize_name($name));
+
+	return $rfc_typemap{$name}
+		if exists $rfc_typemap{$name};
+
+	if(ref $self) {
+		if(my $rule = $self->matchingrule_for_attribute($name, 'equality')) {
+			$rule = lc($rule);
+			return $rfc_typemap{$rule}
+				if exists $rfc_typemap{$rule};
+		}
+		if(my $syntax = $self->attribute_syntax($name)) {
+			$syntax = lc($syntax);
+			return $rfc_typemap{$syntax}
+				if exists $rfc_typemap{$syntax};
+		}
+	}
+
+	return undef;
+}
+
+sub canonicalize_value {
+	my ($self, $name, $value) = @_;
+
+	my $type = $self->canonicalize_type($name);
+
+	unless(defined $type) {
+#		if(ref $self) {
+#			warn "no canonicalization strategy available for '$name' - incomplete schema?\n";
+#		} else {
+#			warn "no canonicalization strategy available for '$name' - consider loading the schema\n";
+#		}
+		return $value;
+	}
+
+	return $self->$type($value);
+}
+
 sub canonicalize_dn {
 	my ($self, $dn) = @_;
 
@@ -1777,17 +2267,31 @@ Net::LDAP::Schema::Canonical - transform LDAP values to a canonical representati
 
 =head1 USAGE
 
-Despite the name, Net::LDAP::Schema::Canonical registers its functions in
-the Net::LDAP::Schema class.
+Loading this package extends Net::LDAP::Schema with three methods, one to
+canonicalize attribute names, one to canonicalize attribute values and one
+to canonicalize DNs.
+
+Canonicalization in this context means: a transformation to a form that is
+the same for any two entries that LDAP would consider equal.
+
+For example, "cn=Bob" and "cn=BOB" are the same name as far as LDAP is
+concerned, but if you do a string comparison they would appear to be
+different. The functions in this package would transform both to "cn=bob",
+so that a string comparison will give the correct result. Unlike a simple
+lc() they resolve all syntactic peculiarities of LDAP, not just the simple
+case map. String escapes, unicode normalization, RDN reordering, etcetera.
+
+Note that despite the name, Net::LDAP::Schema::Canonical registers its
+functions in the Net::LDAP::Schema class.
 
 It does this because schema information is necessary for proper
 canonicalization: different LDAP attribute types have different
 canonicalization rules. Some types are case sensitive, some are not, some
-(such as integers) require further rewriting beyond simple insigificant
+(such as timestamps) require further rewriting beyond simple insigificant
 whitespace removal or case mapping.
 
 By registering its functions in the Net::LDAP::Schema class, they are
-immediately available for use when you have retrieve the LDAP schema.
+immediately available for use when you retrieve the LDAP schema.
 
 However, a schema may not always be available. In such cases you can always
 invoke the canonicalization methods as class methods, i.e.:
@@ -1800,9 +2304,11 @@ instead of:
 
 The methods will then fall back to a small built-in list of common
 attributes and their respective canonicalization rules (to be precise, the
-ones specified in RFC 4519).
+ones specified in RFC 4517/4519).
 
-=head2 $schema->canonicalize_value($name, $value)
+Any errors will cause an exception to be thrown. Use eval {} as necessary.
+
+=head2 $schema->canonicalize_value($name => $value)
 
 Canonicalizes a value based on the type of the attribute name. For example,
 if $name is "cn", a case insensitive canonicalization is applied. Instead
@@ -1811,8 +2317,8 @@ example, caseIgnoreMatch).
 
 =head2 $schema->canonicalize_dn($dn)
 
-Canonicalize a distinguished name (DN) by disassembling the constituent
-RDNs and normalizing each of them according to their respective
+Canonicalize a distinguished name (DN) by disassembling it into its
+constituent RDNs and normalizing each of them according to their respective
 canonicalization rules.
 
 The returned string is guaranteed to be valid UTF-8 but is not marked as
@@ -1823,20 +2329,21 @@ as necessary.
 =head2 $schema->canonicalize_name($name)
 
 Return the canonical name for an attribute. For example, "CN", "commonName"
-and "2.5.4.3" all resolve to "cn". Any options (the bits after the
+and "2.5.4.3" all resolve to "cn". Any options (the bits after a
 semicolon) are lowercased but otherwise left unmodified. For example,
 "OU;lang-EN" becomes "ou;lang-en".
 
 =head1 LIMITATIONS
 
 Invalid values are sometimes flagged as such, but not always. This class is
-not intended as a validator, so the "Garbage In = Garbage Out" (GIGO) rule
+not intended as a validator, so the "Garbage In = Garbage Out" rule
 applies.
 
-Some data types (bit string, boolean, fax numbers, time, lists) are not
-handled at all, or not handled completely. It is the author's intention to
-implement canonicalization for all datatypes specified in RFC 4517 which
-have a Match rule.
+Some data types (bit string, fax numbers) are not handled at all, or not
+handled completely. Trying to canonicalize such values will elicit a
+warning message. Override $SIG{__WARN__} as necessary to catch these.
+It is the author's intention to implement canonicalization for these and
+all other RFC 4517 datatypes which have a Match rule.
 
 Values that are represented using a sharp (#) sign followed by the
 hexadecimal encoded BER representation of their ASN.1 datatype are not
